<html>
<head>
    <title>Upmdcli Media Server {{title}}</title>
    <link rel="stylesheet" type="text/css" href="static/style.css">
    %if reloadsecs:
        <meta http-equiv="refresh" content="{{reloadsecs}}">
    %end
</head>
<body>
