# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class ArtistIdName:

    def __init__(
            self,
            artist_id: str,
            artist_name: str):
        self.__artist_id: str = artist_id
        self.__artist_name: str = artist_name

    @property
    def artist_id(self) -> str:
        return self.__artist_id

    @property
    def artist_name(self) -> str:
        return self.__artist_name


class ArtistIdNameCoverArt(ArtistIdName):

    def __init__(
            self,
            artist_id: str,
            artist_name: str,
            cover_art: str = None):
        super().__init__(
            artist_id=artist_id,
            artist_name=artist_name)
        self.__cover_art: str = cover_art

    @property
    def cover_art(self) -> str:
        return self.__cover_art
