body {    margin: 0; font-family: sans-serif }
* { font-family: inherit  }
#searchbox {
    background: #eee;
    border-bottom: 1px solid #666;
}

#status {
    margin: 1em;
    background: #fcfcfc;
    border: 1px solid #eaeaea;
    color: #666;
    font-size: 9pt;
}
#found { float: left;     padding: 0.5em; }
#downloads { float: right; border-left: 1px solid #eaeaea; padding: 0.5em } 
#downloads a {
    padding: 5px;
    color: #999;
    font-size: 7pt;
}
#downloads a:hover { background: #ccc; color: white  }

#pages { clear: both; float: none; width: 100%; text-align: center; }
.page {
    font-size: 8pt;
    color: gray;
    padding-left: 6px; padding-right: 6px;
    padding-top: 2px; padding-bottom: 2px;
    border: 1px solid #ddd;
    background: #f5f5f5;

}
.page:hover, .current { background: #ccc; color: white; }

#results { padding: 1em; }
.search-result {
    margin-left: 15%;
    margin-right: 15%;
    background: #fff;
    padding-bottom: 0.5em;
    padding-top: 1em;
}
.search-result-title, .search-result-ipath {
    margin-bottom: 6px;
    font-weight: bold;
    font-size: 12pt;
    float: left;
}
.search-result-ipath { color: #aaa; padding-left: 1em; }
.search-result-title a { text-decoration: none; color: black; }
.search-result-author {
    clear: both;
    color: #666;
    font-size: 7pt;
    margin-bottom: 6px
}
.search-result-number {
    width: 4em;
    float: left;
    margin-left: -5em;
    text-align: right;
}
.search-result-number a { font-size: 10pt; font-weight: normal; color:#bbb; background: #f8f8f8; padding: 4px; }
.search-result-number a:hover { color:white; background: #ccc; }
.search-result-snippet {
    margin-left: 1px;
    margin-top: 2px;
    font-size: 10pt;
    text-align: justify;
    border-left: 3px solid #ddd;
    padding-left: 1em;
    color: #444; clear: both;
    font-family: serif;
}
.search-result-url { clear:left; }
.search-result-url a { color: #5a5; font-size: 8pt; float: left; margin-right: 1em;}
.search-result-links { color: #aaf; font-size: 8pt; float: left; margin-right: 1em; visibility: hidden }
.search-result-links a { padding-right: 0.5em; }
.search-result-links a:hover { text-decoration: underline }
.search-result:hover > .search-result-links { visibility: visible }
.search-result-date { color: #777; font-size: 8pt; float: right; margin-bottom: 7px; }
.search-result-size { color: #777; font-size: 8pt; float: right; display: none; }
.search-result-highlight { color: #7E1212; font-weight: bold; }
.gray { color: #aaa }

a { text-decoration: none }
table td {
    padding: 10;
    border-right: 1px dotted black ;
    border-collapse: collapse;
}
table td:last-child {
    border-right: 0px;
}
form { margin: 0 }
input, select {
    margin-top: 2px;
    margin-bottom: 2px;
    background: white;
    border: 1px solid #999;
    padding: 3px;
    width: 100%;
    color: #555;
}
input:focus, button:focus, select:focus{
    border: 1px solid black;
    color: black;
}
input[type=button],input[type=reset],input[type=submit] { width: 32%; }
input[type=button]:hover,input[type=reset]:hover,input[type=submit]:hover{ color:white; background: #ccc; cursor: pointer; }
input[type=reset]:hover,input[type=submit]:hover { }
input[name=before], input[name=after] { width: 45% }
tr { vertical-align: top }

#settings-box {
    padding: 1em;
    position: absolute;
    top: 5%;
    left: 30%;
    right: 30%;
    background: #eee;
    border: 1px solid #999;
}
hr { border: 1px solid #999 }

#fade {
    position: absolute;
    width: 100%;
    z-index: 1;
    opacity: 0.7;
    background: white;
    display: none;
}

